#! /usr/bin/env python3

import argparse
import csv
from functools import reduce
import json
from pathlib import Path
from typing import Dict, List, Union

import matplotlib.pyplot as plt
import numpy as np

from emod_api.channelreports.utils import read_json_file, get_report_channels, accumulate_trace_data, save_to_csv, plot_traces
from emod_api.channelreports.utils import _validate_property_report_channels, _validate_property_report_ips


def main(args: argparse.Namespace):

    jsonData = read_json_file(args.filename)
    poolData = get_report_channels(jsonData)
    poolKeys = sorted(poolData.keys())

    if args.verbose:
        print("Channels:Pools-")
        print(json.dumps(poolKeys, indent=4))

    if args.list:
        listChannelsAndIPs(poolKeys)
        return

    _validate_property_report_channels(args.channels, poolData)
    _validate_property_report_ips(args.rollup, poolData)

    (traceValues, normValues) = accumulate_trace_data(args.channels, args.verbose, args.rollup, args.by, poolData)

    if args.csv is None:
        plotTraces(args, traceValues, normValues)
    else:
        save_to_csv(traceValues, normValues, args.csv, args.normalize, args.transpose)

    return


def listChannelsAndIPs(channelKeys: List[str]) -> None:

    channels = sorted(set([key.split(":",1)[0] for key in channelKeys]))    # keys look like "CHANNEL:IP:value,...,IP:value"

    print("\nChannels:")
    for channel in channels:
        print(f"\t{channel}")

    # Each channel _should_ have the same set of IPs, but we'll check them all
    csvkvps = [key.split(":",1)[1] for key in channelKeys]              # For each channel get a comma separated list of IP:value pairs (see format above)
    kvplists = [csv.split(",") for csv in csvkvps]                      # For each CSV convert to actual list by splitting on ","
    ips = [map(lambda t: t.split(":")[0], kvps) for kvps in kvplists]   # Convert each IP:value entry to just IP
    properties = sorted(reduce(lambda s, e: s.union(e), ips, set()))    # Add all IPs to an initially empty set

    print("\nIPs:")
    for property in properties:
        print(f"\t{property}")

    print()

    return


def plotTraces(args: argparse.Namespace, traceValues: Dict[str, np.ndarray], normValues: Union[int, np.ndarray]) -> None:

    if args.verbose:
        print(sorted(traceValues.keys())) if args.verbose else None

    figure = plot_traces(traceValues, normValues, args.overlay, args.channels, args.normalize, args.filename, args.legend, args.by)

    if args.saveFigure:
        print("Saving figure 'propertyReport.png'...")
        figure.savefig('propertyReport.png')

    plt.show()

    return


def processCommandline() -> argparse.Namespace:

    parser = argparse.ArgumentParser(description='Property Report Plotting')
    parser.add_argument('filename', nargs='?', default='PropertyReport.json', help='property report filename [PropertyReport.json]')
    parser.add_argument('-c', '--channel', action='append', help='channel(s) to display [Infected]', metavar='channelName', dest='channels')
    parser.add_argument('-r', '--rollup', action='append', help="IP(s) under which to roll up other IP keys and values")
    parser.add_argument('-n', '--normalize', help='plot channel(s) normalized by statistical population', action='store_true')
    parser.add_argument('-b', '--by', default="Statistical Population", help="Channel for normalization ['Statistical Population']")
    parser.add_argument('-o', '--overlay', help='overlay pools of the same channel', action='store_true')
    parser.add_argument('-s', '--save', help="save figure to file 'propertyReport.png'", action='store_true', dest='saveFigure')
    parser.add_argument('-v', '--verbose', action="store_true")
    parser.add_argument('--no-legend', action="store_false", dest="legend")     # Note args.legend default to True, passing --no-legend sets args.legend to False
    parser.add_argument('-l', '--list', action="store_true", help="List channels and IP keys found in the report. No plotting is performed with this option.")
    parser.add_argument("--csv", type=Path, default=None, help="Write data for selected channel(s) to given file.")
    parser.add_argument("-t", "--transpose", action="store_true", help="write channels as columns rather than rows (only in effect with '--csv' option)")

    args = parser.parse_args()

    if not args.channels:
        args.channels = ['Infected']

    if args.rollup is not None and len(args.rollup) == 1 and args.rollup[0].lower() == "all":
        args.rollup = []

    if not args.list:
        print(f"Filename:              '{args.filename}'")
        print(f"Channel(s):            {args.channels}")
        print(f"Rollup:                {args.rollup}")
        print(f"Normalize:             {args.normalize}")
        print(f"Normalization Channel: '{args.by}'") if args.normalize else None
        print(f"Overlay:               {args.overlay}")
        print(f"Save:                  {args.saveFigure}")
        print(f"CSV filename:          '{args.csv}'") if args.csv else None
        print(f"Transpose CSV:         {args.transpose}")

    return args


if __name__ == '__main__':

    args = processCommandline()
    main(args)
